#define F_CPU 16000000

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <math.h>

#include "spi.h"
#include "pm.h"


#define F_AUDIO 18000 //18000
#define F_LONG_TIME 20// this has to be a divisor of F_AUDIO and 1000
#define LONG_TIME_MS (1000 / F_LONG_TIME) // long timestep in ms
#define LONG_TIME_Q (F_AUDIO / F_LONG_TIME) // number of interrupts necesary to increment long_time
#define DEBOUNCE_TIME_MS 50 // debounce time. Accuracy depends on LONG_TIME_MS
#define DEBOUNCE_TIME_Q (DEBOUNCE_TIME_MS / LONG_TIME_MS) // delta long_time necesary for debouncing. This has to end up smaller than 255

#define BUTTONS_N 16 // has to be multiple of 8
#define BUTTONS_KEYBOARD 13
#define BUTTON_OCT_DOWN 15
#define BUTTON_OCT_UP 14
#define B_STATE_OFF 0
#define B_STATE_PUSHED 1
#define B_STATE_ON 2
#define B_STATE_RELEASED 3
#define B_STATE_PRE_PUSHED 4
#define B_STATE_PRE_RELEASED 5


#define SOUND_SAMPLES 256 // AFFECTS MIPMAP_LENGTH. has to be a power of 2
#define MIPMAP_LENGTH (256+128+64+32+16+8 + NOTE_OCTAVES+1)  //For every octave, half of the previous SOUND_SAMPLES is added + 1
#define SOUND_BASE_F ((double)F_AUDIO / SOUND_SAMPLES)
#define SOUND_VOICES 4

#define NOTE_BASE_F SOUND_BASE_F
#define NOTE_OCTAVES 5 // AFFECTS MIPMAP_LENGTH
#define NOTES_N (NOTE_OCTAVES * 12 + 1)
#define START_OCTAVE 2

#define KEY_NO_VOICE 255

#define V_STATE_OFF 0
#define V_STATE_HIT 1
#define V_STATE_SUSTAIN 2
#define V_STATE_RELEASE 3

#define ENVELOPE_HIT_SAMPLES 140 //strictly less than 256
#define ENVELOPE_HIT_TIME ((float)ENVELOPE_HIT_SAMPLES / F_LONG_TIME)
#define ENVELOPE_RELEASE_SAMPLES 140 //strictly less than 256


//unsigned short long_time_i = 0;
volatile unsigned char long_time = 0;

//notes
note_t notes[NOTES_N];

//buttons
unsigned char button_state[BUTTONS_N];
unsigned char button_debounce_count[BUTTONS_N];
key_t keys[BUTTONS_KEYBOARD];
unsigned char octave_start = START_OCTAVE * 12; //index of the first note of this octave from notes

//voices
signed char wave_mipmaps[MIPMAP_LENGTH] = {0};
volatile voice_t voices[SOUND_VOICES];
volatile unsigned char voices_on = 0;

//envelopes
unsigned char env_hit_end;
unsigned char env_release_end;

unsigned char env_amp_hit[ENVELOPE_HIT_SAMPLES];
unsigned char env_amp_sustain;
unsigned char env_amp_release[ENVELOPE_RELEASE_SAMPLES]; //the values represent what to substract from the amplitude registered at the moment of release (0 = full, 255 = min);

unsigned char env_lpf_hit[ENVELOPE_HIT_SAMPLES]; //0 is low cutoff frequency, 255 is high
unsigned char env_lpf_sustain;
unsigned char env_lpf_release[ENVELOPE_RELEASE_SAMPLES];
unsigned char env_lpf_release_offset;

signed short signal = 0;
//unsigned char int_test = 0;

ISR(TIMER1_COMPA_vect)
{
	long_time ^= 1;
}

ISR(TIMER0_COMPA_vect)
{
	/*if(int_test == 1)
		PORTD |= 1 << PD7;
	sei();
	int_test = 1;*/
	
	//calculate signal to send
	unsigned short signal_send = signal + 2048;
	unsigned char high_byte_send = 64 | ((signal_send >> 8) & 15);
	
	//send signal
	PORTB &= ~(1 << PB4);
	SPDR = high_byte_send;	
	unsigned char low_byte_send = signal_send & 255;

	signal = 0;
	
	while(!(SPSR & (1 << SPIF)));
	SPDR = low_byte_send;
	
	for (unsigned char i = 0; i != SOUND_VOICES; ++i) {
		if (voices[i].state != V_STATE_OFF) {
			//interpolation
			signed short amplitude = (wave_mipmaps[voices[i].mipmap_current] * (256 - voices[i].sample_overstep) + wave_mipmaps[voices[i].mipmap_current+1] * voices[i].sample_overstep) / 256;
			//lpf
			amplitude = voices[i].last_value + ((amplitude - voices[i].last_value) * (voices[i].env_lpf_multiplier + 1)) / 256;
			voices[i].last_value = amplitude;
			//amp
			amplitude = (amplitude * (voices[i].env_amp_multiplier+1)) / 256;
			
			signal += amplitude;
			
			if (voices[i].sample_overstep > 255 - voices[i].sample_overstep_inc) {
				voices[i].mipmap_current += 2;
			} else {
				voices[i].mipmap_current++;
			}
			
			voices[i].sample_overstep += voices[i].sample_overstep_inc;
			if (voices[i].mipmap_current >= voices[i].mipmap_end) {
				voices[i].mipmap_current -= voices[i].mipmap_length;
			}
			
		}
	}
	
	signal *= (4096 / (SOUND_VOICES * 256));
	while(!(SPSR & (1 << SPIF)));
	PORTB |= (1 << PB4);
	
	//int_test = 0;
}

void compute_envelopes()
{
	//1 sec rise to sustain && 1 sec release
	float val = 255;
	unsigned char i;
	for (i = 0; val >= 50; ++i, val -= (float)200 / (F_LONG_TIME)) {
		//env_amp_hit[i] = (unsigned char) val;
		env_amp_hit[i] = 255;
		env_lpf_hit[i] = (unsigned char) (val);
		//env_lpf_hit[i] = 255;
	}
	env_hit_end = i;
	env_amp_sustain = 255;
	env_lpf_sustain = 50;
	val = 255;
	for (i = 0; val >= 0; ++i, val -= (float)256 / (F_LONG_TIME*2)) {
		//env_amp_release[i] = 255;
		env_amp_release[i] = val;
		env_lpf_release[i] = val;
		//env_lpf_release[i] = 255;
	}
	env_release_end = i;
	env_lpf_release_offset = 50;
	
	//flat
	/*env_amp_sustain = 255;
	env_lpf_sustain = 255;
	env_amp_hit[0] = 255;
	env_lpf_hit[0] = 255;
	env_hit_end = 1;
	env_amp_release[0] = 0;
	env_lpf_release[0] = 0;
	env_release_end = 1;
	env_lpf_release_offset = 0;*/	
}

void generate_mipmaps()
{
	unsigned short mipmap_start = SOUND_SAMPLES+1;
	unsigned short mipmap_length = SOUND_SAMPLES / 2;
	unsigned short mipmap_end = mipmap_start + mipmap_length;
	unsigned char current_octave = 0;
	
	/*unsigned char step = 2;
	
	while (current_octave <= NOTE_OCTAVES) {
		for (unsigned short i = mipmap_start, j = 0; i < mipmap_start + mipmap_length; i++, j += step) {
			wave_mipmaps[i] = wave_mipmaps[j];
		}
		wave_mipmaps[mipmap_start+mipmap_length] = wave_mipmaps[0];
		
		mipmap_start += mipmap_length+1;
		mipmap_length /= 2;
		step *= 2;
		current_octave++;
	}*/
	
	unsigned short prev_mipmap_start = 0;
	unsigned short prev_mipmap_length = SOUND_SAMPLES;
	unsigned short prev_mipmap_end = SOUND_SAMPLES;

	float kernel[3] = {0.5, 1, 0.5};
	float kernel_div = 2;
	float sum;
	for (unsigned char k = 0; k <= NOTE_OCTAVES; ++k) {
		//first sample
		sum = 0;
		sum += kernel[0] * wave_mipmaps[prev_mipmap_end - 1];
		sum += kernel[1] * wave_mipmaps[prev_mipmap_start];
		sum += kernel[2] * wave_mipmaps[prev_mipmap_start + 1];
		wave_mipmaps[mipmap_start] = sum / kernel_div;
		
		//last sample
		sum = 0;
		sum += kernel[0] * wave_mipmaps[prev_mipmap_end-2];
		sum += kernel[1] * wave_mipmaps[prev_mipmap_end-1];
		sum += kernel[2] * wave_mipmaps[prev_mipmap_start];
		wave_mipmaps[mipmap_end-1] = sum / kernel_div;
		
		//middle
		for (unsigned short i = mipmap_start+1, j = prev_mipmap_start+2; i < mipmap_end - 1; ++i, j += 2) {
			sum = 0;
			sum += kernel[0] * wave_mipmaps[j-1];
			sum += kernel[1] * wave_mipmaps[j];
			sum += kernel[2] * wave_mipmaps[j+1];
			wave_mipmaps[i] = sum / kernel_div;
		}
		
		wave_mipmaps[mipmap_end] = wave_mipmaps[mipmap_start];
		
		prev_mipmap_start = mipmap_start;
		prev_mipmap_length = mipmap_length;
		prev_mipmap_end = mipmap_end;
		
		mipmap_start += mipmap_length + 1;
		mipmap_length /= 2;
		mipmap_end = mipmap_start + mipmap_length;
	}
}

void compute_waveform()
{	
	//square
	/*for (unsigned short i = 0; i < SOUND_SAMPLES; i++) {
		if (i < SOUND_SAMPLES / 2) {
			wave_mipmaps[i] = -128;
		} else {
			wave_mipmaps[i] = 127;
		}
	}*/
	
	//saw
	/*signed char val = -128;
	for (unsigned short i = 0; i < SOUND_SAMPLES; ++i, val++) {
		wave_mipmaps[i] = val;
	}*/
	
	//triangle
	for (unsigned short i = 0, val = -128; i < SOUND_SAMPLES / 2; i++, val += 256 / (SOUND_SAMPLES / 2)) {
		wave_mipmaps[i] = val;
	}
	for (unsigned short i = SOUND_SAMPLES / 2, val = 127; i < SOUND_SAMPLES; i++, val -= 256 / (SOUND_SAMPLES / 2)) {
		wave_mipmaps[i] = val;
	}
	
	//sinus
	/*double w = 2 * M_PI / 256;
	for (unsigned short i = 0; i < SOUND_SAMPLES; i++) {
		double temp = sin(w*i);
		wave_mipmaps[i] = temp * 126;
	}*/
	
	
	wave_mipmaps[SOUND_SAMPLES] = wave_mipmaps[0];
}

void init_notes()
{
	const double a = 1.059463094359;
	double a_n = 1;
	double freq;
	
	unsigned short mipmap_start = 0;
	unsigned short mipmap_length = SOUND_SAMPLES;
	double mipmap_freq = NOTE_BASE_F;
	
	for (unsigned char i = 0; i < NOTES_N; i++) {
		freq = NOTE_BASE_F * a_n;
		
		if (freq >= mipmap_freq*2) {
			mipmap_start += mipmap_length+1;
			mipmap_length /= 2;
			mipmap_freq *= 2;
		}
		
		notes[i].mipmap_start = mipmap_start;
		notes[i].mipmap_length = mipmap_length;
		notes[i].sample_overstep_inc = (freq / mipmap_freq - 1) * 256;
		
		a_n *= a;
	}
}

void init_voices()
{
	for (unsigned char i = 0; i < SOUND_VOICES; i++) {
		voices[i].state = V_STATE_OFF;
		voices[i].key_index = BUTTONS_N; //a number outside the range
	}
}

void init_buttons()
{
	//all buttons
	for (unsigned char i = 0; i < BUTTONS_N; i++) {
		button_debounce_count[i] = 0;
		button_state[i] = B_STATE_OFF;
	}
	
	//keys
	for (unsigned char i = 0; i < BUTTONS_KEYBOARD; i++) {
		keys[i].voice_index = KEY_NO_VOICE;
	}
}

void init_regs()
{
	float aux_float;
	
	// led
	DDRD |= 1 << PD7;
	
	// buttons
	DDRA = 0;
	DDRC = 0;
	PORTA = 255;
	PORTC = 255;
	
	// SPI
	SPI_init();
	
	// init timer 1
	/*float aux_float = (float)F_CPU / F_AUDIO + 0.5f;
    OCR1A = (short)aux_float; 
    TIMSK1 = (1 << OCIE1A); // enable OCR1A interrupt
    TCCR1B = (1 << CS10) | (1 << WGM12); // OCR1A CTC & prescaler = 1*/
	
	//audio timer
	aux_float = (float) F_CPU / 8 / F_AUDIO + 0.5f;
	OCR0A = (unsigned char)aux_float;
	TCCR0A = 1 << WGM01;
	TCCR0B = (2 << CS00); // prescaler = 8
	TIMSK0 = 1 << OCIE0A;
	
	//long time timer
	aux_float = (float) F_CPU / 64 / F_LONG_TIME + 0.5f;
	OCR1A = (unsigned short) aux_float;
	TIMSK1 = (1 << OCIE1A);
	TCCR1B = (3 << CS10) | (1 << WGM12); //OCR1A CTC & prescaler = 64;
	
}

int main(void)
{
	init_regs();
	init_notes();
	init_buttons();
	init_voices();
	
	compute_waveform();
	generate_mipmaps();
	compute_envelopes();
	
	unsigned char local_long_time = long_time;
	
	sei();
	
	for (;;) {
		//decrement debounce counter
		if (long_time != local_long_time) { // WARNING: in theory, the difference should always be 1
			local_long_time = long_time;
			for (unsigned char i = 0; i < BUTTONS_N; i++) {
				if (button_debounce_count[i] != 0) {
					button_debounce_count[i]--;
				}
			}
			//update envelopes
			for (unsigned char i = 0; i != SOUND_VOICES; ++i) {
				if (voices[i].state == V_STATE_HIT) {
					voices[i].env_index++;
					if (voices[i].env_index != env_hit_end) {
						voices[i].env_amp_multiplier = env_amp_hit[voices[i].env_index];
						voices[i].env_lpf_multiplier = env_lpf_hit[voices[i].env_index];
					} else {
						voices[i].state = V_STATE_SUSTAIN;
						voices[i].env_amp_multiplier = env_amp_sustain;
						voices[i].env_lpf_multiplier = env_lpf_sustain;
					}
				} else if (voices[i].state == V_STATE_RELEASE) {
					voices[i].env_index++;
					if (voices[i].env_index == env_release_end) {
						voices[i].state = V_STATE_OFF;
						voices_on--;
					} else {
						voices[i].env_amp_multiplier = (env_amp_release[voices[i].env_index] * (voices[i].env_amp_release_base+1)) / 256;
						voices[i].env_lpf_multiplier = voices[i].env_lpf_release_target + (env_lpf_release[voices[i].env_index] * (voices[i].env_lpf_release_scale+1)) / 256;
					}
				}
			}
		}
		
		//button state update ******************
		for (unsigned char i = 0; i < 8; i++) {
			if (button_state[i] == B_STATE_OFF && (PINA & (1 << i)) == 0) {
				button_state[i] = B_STATE_PRE_PUSHED;
				button_debounce_count[i] = DEBOUNCE_TIME_Q;
			} else if (button_state[i] == B_STATE_PRE_PUSHED) {
				if ((PINA & (1 << i)) != 0) {
					button_state[i] = B_STATE_OFF;
					button_debounce_count[i] = 0;
				} else if (button_debounce_count[i] == 0) {
					button_state[i] = B_STATE_PUSHED;
				}
			} else if (button_state[i] == B_STATE_PUSHED) {
				button_state[i] = B_STATE_ON;
			} else if (button_state[i] == B_STATE_ON && (PINA & (1 << i)) != 0) {
				button_state[i] = B_STATE_PRE_RELEASED;
				button_debounce_count[i] = DEBOUNCE_TIME_Q;
			} else if (button_state[i] == B_STATE_PRE_RELEASED) {
				if ((PINA & (1 << i)) == 0) {
					button_state[i] = B_STATE_ON;
					button_debounce_count[i] = 0;
				} else if (button_debounce_count[i] == 0) {
					button_state[i] = B_STATE_RELEASED;
				}
			} else if (button_state[i] == B_STATE_RELEASED) {
				button_state[i] = B_STATE_OFF;
			}
		}
		
		for (unsigned char i = 8; i < 16; i++) {
			if (button_state[i] == B_STATE_OFF && (PINC & (1 << (i-8))) == 0) {
				button_state[i] = B_STATE_PRE_PUSHED;
				button_debounce_count[i] = DEBOUNCE_TIME_Q;
			} else if (button_state[i] == B_STATE_PRE_PUSHED) {
				if ((PINC & (1 << (i-8))) != 0) {
					button_state[i] = B_STATE_OFF;
					button_debounce_count[i] = 0;
				} else if (button_debounce_count[i] == 0) {
					button_state[i] = B_STATE_PUSHED;
				}
			} else if (button_state[i] == B_STATE_PUSHED) {
				button_state[i] = B_STATE_ON;
			} else if (button_state[i] == B_STATE_ON && (PINC & (1 << (i-8))) != 0) {
				button_state[i] = B_STATE_PRE_RELEASED;
				button_debounce_count[i] = DEBOUNCE_TIME_Q;
			} else if (button_state[i] == B_STATE_PRE_RELEASED) {
				if ((PINC & (1 << (i-8))) == 0) {
					button_state[i] = B_STATE_ON;
					button_debounce_count[i] = 0;
				} else if (button_debounce_count[i] == 0) {
					button_state[i] = B_STATE_RELEASED;
				}
			} else if (button_state[i] == B_STATE_RELEASED) {
				button_state[i] = B_STATE_OFF;
			}
		}
		//*****************
		
		for (unsigned char i = 0; i < BUTTONS_KEYBOARD; i++) {
			if (button_state[i] == B_STATE_PUSHED) {
				unsigned char voice_place = 255;
				
				if (voices_on < SOUND_VOICES) {
					for (unsigned char j = 0; j < SOUND_VOICES; j++) {
						if (voices[j].state == V_STATE_OFF) {
							voice_place = j;
						} else if (voices[j].key_index == i && voices[j].state == V_STATE_RELEASE) {
							voice_place = j;
							voices[voice_place].state = V_STATE_OFF;
							voices_on--;
							break;
						}
					}
				} else {
					unsigned char max_env_index = 0;
					for (unsigned char j = 0; j < SOUND_VOICES; j++) {
						if (voices[j].state == V_STATE_RELEASE) {
							if (voices[j].key_index == i) {
								voice_place = j;
								break;
							} else if (voices[j].env_index >= max_env_index) {
								voice_place = j;
								max_env_index = voices[j].env_index;
							}
						}
					}
					
					if (voice_place != 255) {
						voices[voice_place].state = V_STATE_OFF;
						voices_on--;
					} else goto do_nothing;
				}
				
				voices[voice_place].mipmap_start = notes[octave_start + i].mipmap_start;
				voices[voice_place].mipmap_end = notes[octave_start + i].mipmap_start + notes[octave_start + i].mipmap_length;
				voices[voice_place].mipmap_length = notes[octave_start + i].mipmap_length;
				voices[voice_place].sample_overstep_inc = notes[octave_start + i].sample_overstep_inc;
				
				voices[voice_place].sample_overstep = 0;
				voices[voice_place].mipmap_current = notes[octave_start + i].mipmap_start;
				
				voices[voice_place].env_index = 0;
				voices[voice_place].env_amp_multiplier = env_amp_hit[0];
				voices[voice_place].env_lpf_multiplier = env_lpf_hit[0];
				voices[voice_place].last_value = 0;
				
				keys[i].voice_index = voice_place;
				voices[voice_place].key_index = i;
				voices[voice_place].state = V_STATE_HIT;
				voices_on++;
				
				do_nothing:;
			} else if (button_state[i] == B_STATE_RELEASED && keys[i].voice_index != KEY_NO_VOICE) {
				voices[keys[i].voice_index].env_amp_release_base = voices[keys[i].voice_index].env_amp_multiplier;
				voices[keys[i].voice_index].env_index = 0;
				
				unsigned char env_lpf_release_base = voices[keys[i].voice_index].env_lpf_multiplier;
				if (env_lpf_release_base >= env_lpf_release_offset) {
					voices[keys[i].voice_index].env_lpf_release_scale = env_lpf_release_base - env_lpf_release_offset;
					voices[keys[i].voice_index].env_lpf_release_target = env_lpf_release_offset;
				} else {
					voices[keys[i].voice_index].env_lpf_release_scale = 0;
					voices[keys[i].voice_index].env_lpf_release_target = env_lpf_release_base;
				}
				
				voices[keys[i].voice_index].state = V_STATE_RELEASE;
				keys[i].voice_index = KEY_NO_VOICE;
			}
		}
		
		if (button_state[BUTTON_OCT_DOWN] == B_STATE_PUSHED) {
			if (octave_start > 0) {
				octave_start -= 12;
				for (unsigned char i = 0; i < BUTTONS_KEYBOARD; i++) {
					if (keys[i].voice_index != KEY_NO_VOICE) {
						voices[keys[i].voice_index].env_amp_release_base = voices[keys[i].voice_index].env_amp_multiplier;
						voices[keys[i].voice_index].env_index = 0;
						
						unsigned char env_lpf_release_base = voices[keys[i].voice_index].env_lpf_multiplier;
						if (env_lpf_release_base >= env_lpf_release_offset) {
							voices[keys[i].voice_index].env_lpf_release_scale = env_lpf_release_base - env_lpf_release_offset;
							voices[keys[i].voice_index].env_lpf_release_target = env_lpf_release_offset;
						} else {
							voices[keys[i].voice_index].env_lpf_release_scale = 0;
							voices[keys[i].voice_index].env_lpf_release_target = env_lpf_release_base;
						}
						
						voices[keys[i].voice_index].state = V_STATE_RELEASE;
						keys[i].voice_index = KEY_NO_VOICE;
					}
				}
			}
		} else if (button_state[BUTTON_OCT_UP] == B_STATE_PUSHED) {
			if (octave_start < (NOTE_OCTAVES-1)*12) {
				octave_start += 12;
				for (unsigned char i = 0; i < BUTTONS_KEYBOARD; i++) {
					if (keys[i].voice_index != KEY_NO_VOICE) {
						voices[keys[i].voice_index].env_amp_release_base = voices[keys[i].voice_index].env_amp_multiplier;
						voices[keys[i].voice_index].env_index = 0;
						
						unsigned char env_lpf_release_base = voices[keys[i].voice_index].env_lpf_multiplier;
						if (env_lpf_release_base >= env_lpf_release_offset) {
							voices[keys[i].voice_index].env_lpf_release_scale = env_lpf_release_base - env_lpf_release_offset;
							voices[keys[i].voice_index].env_lpf_release_target = env_lpf_release_offset;
						} else {
							voices[keys[i].voice_index].env_lpf_release_scale = 0;
							voices[keys[i].voice_index].env_lpf_release_target = env_lpf_release_base;
						}
						
						voices[keys[i].voice_index].state = V_STATE_RELEASE;
						keys[i].voice_index = KEY_NO_VOICE;
					}
				}
			}
		}
		
	}
		
	return 0;
}
